//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <values.h>
#include "MySysUtils.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
namespace
{
const int BITSPERDWORD = BITSPERBYTE * sizeof(DWORD);
const DWORD FULLDWORD = static_cast<DWORD>(-1);
} // end of namespace
//---------------------------------------------------------------------------
// control utilities
//---------------------------------------------------------------------------
void __fastcall SaveListItemsToFile(AnsiString AFileName,
	const TListView* AListView, bool AIncludeColumns)
{
	TStringList* lines = new TStringList();
	TStringList* line = new TStringList();
	try {
		lines->Capacity = (AIncludeColumns ? 1 : 0) + AListView->Items->Count;
		if(AIncludeColumns) {
			line->Clear();
			for(int i = 0; i < AListView->Columns->Count; i++) {
				line->Add(AListView->Columns->Items[i]->Caption);
			}
			lines->Add(line->CommaText);
		}
		for(int i = 0; i < AListView->Items->Count; i++) {
			line->Clear();
			line->Add(AListView->Items->Item[i]->Caption);
			line->AddStrings(AListView->Items->Item[i]->SubItems);
			while(line->Count < AListView->Columns->Count) {
				line->Add("");
			}
			while(line->Count > AListView->Columns->Count) {
				line->Delete(lines->Count - 1);
			}
			lines->Add(line->CommaText);
		}
		lines->SaveToFile(AFileName);
	}
	__finally {
		delete line;
		delete lines;
	}
}
//---------------------------------------------------------------------------
// file utilities
//---------------------------------------------------------------------------
AnsiString __fastcall ShortFileNameToLongFileName(AnsiString AShortFileName)
{
	AnsiString longFileName = AShortFileName;
	const int bufferSize = MAX_PATH;
	char buffer[bufferSize];
	int resultSize = GetLongPathName(AShortFileName.c_str(), buffer,
		bufferSize);
	if(resultSize > 0 && resultSize < bufferSize) {
		longFileName = buffer;
	}
	return longFileName;
}
//---------------------------------------------------------------------------
bool __fastcall ShowFileProperties(AnsiString AFileName)
{
	bool succeeded = false;
	SHELLEXECUTEINFO params;
	ZeroMemory(&params, sizeof(SHELLEXECUTEINFO));
	params.cbSize = sizeof(SHELLEXECUTEINFO);
	params.lpFile = AFileName.c_str();
	params.lpVerb = "properties";
	params.fMask = SEE_MASK_INVOKEIDLIST | SEE_MASK_FLAG_NO_UI;
	if(ShellExecuteEx(&params) != 0) {
		succeeded = true;
	}
	return succeeded;
}
//---------------------------------------------------------------------------
bool __fastcall MatchFileName(AnsiString AFileName, AnsiString AMask)
{
	bool matched = false;
	const char* namePos = AFileName.c_str();
	const char* maskPos = AMask.c_str();
	TStack* nameStack = new TStack();
	TStack* maskStack = new TStack();
	bool anyChar;
	bool charMatched;
	for(;;) {
		anyChar = false;
		for(;;) {
			if(*maskPos == '?') {
				if(*namePos == '\0') {
					goto EXIT;
				}
				namePos = CharNext(namePos);
			}
			else if(*maskPos == '*') {
				anyChar = true;
			}
			else {
				break;
			}
			maskPos = CharNext(maskPos);
		}
		for(;;) {
			if(IsDBCSLeadByte(*namePos)) {
				charMatched = strncmp(namePos, maskPos, 2) == 0;
			}
			else {
				charMatched = toupper(*namePos) == toupper(*maskPos);
			}
			if(charMatched) {
				if(*namePos == '\0') {
					matched = true;
					goto EXIT;
				}
				if(anyChar) {
					nameStack->Push(const_cast<char*>(CharNext(namePos)));
					maskStack->Push(const_cast<char*>(maskPos));
				}
				namePos = CharNext(namePos);
				maskPos = CharNext(maskPos);
				break;
			}
			if(anyChar && *namePos != '\0') {
				namePos = CharNext(namePos);
				continue;
			}
			if(nameStack->Count() > 0) {
				namePos = static_cast<char*>(nameStack->Pop());
				maskPos = static_cast<char*>(maskStack->Pop());
				anyChar = true;
				continue;
			}
			goto EXIT;
		}
	} EXIT:;
	delete maskStack;
	delete nameStack;
	return matched;
}
//---------------------------------------------------------------------------
AnsiString __fastcall ExtractFileLink(AnsiString AFileName)
{
	AnsiString link;
	IShellLink* iShellLink = NULL;
	IPersistFile* iPersistFile = NULL;
	for(;;) {
		if(FAILED(CoCreateInstance(CLSID_ShellLink, NULL,
			CLSCTX_INPROC_SERVER, IID_IShellLink,
			reinterpret_cast<void**>(&iShellLink)))) {
			iShellLink = NULL;
			break;
		}
		if(FAILED(iShellLink->QueryInterface(IID_IPersistFile,
			reinterpret_cast<void**>(&iPersistFile)))) {
			iPersistFile = NULL;
			break;
		}
		WideString wideFileName = AFileName;
		if(FAILED(iPersistFile->Load(wideFileName.c_bstr(), STGM_READ))) {
			break;
		}
		const int bufferSize = MAX_PATH;
		char buffer[bufferSize];
		WIN32_FIND_DATA data;
		if(FAILED(iShellLink->GetPath(buffer, bufferSize, &data,
			SLGP_UNCPRIORITY))) {
			break;
		}
		link = buffer;
		break;
	}
	if(iPersistFile != NULL) {
		iPersistFile->Release();
	}
	if(iShellLink != NULL) {
		iShellLink->Release();
	}
	return link;
}
//---------------------------------------------------------------------------
// string utilities
//---------------------------------------------------------------------------
AnsiString __fastcall FormatInt(AnsiString ASrc)
{
	const char separator = ',';
	const int group = 3;
	AnsiString dst = ASrc;
	for(;;) {
		int signLen = 0;
		switch(*dst.c_str()) {
			case '-':
			case '+': signLen = 1; break;
		}
		int groupCount = ((dst.Length() - signLen) - 1) / group;
		if(groupCount <= 0) {
			break;
		}
		dst.SetLength(dst.Length() + groupCount);
		char* dstPos = dst.c_str() + dst.Length() - 1;
		char* srcPos = dstPos - groupCount;
		for(int i = 0; i < groupCount; i++) {
			for(int j = 0; j < group; j++) {
				*dstPos-- = *srcPos--;
			}
			*dstPos-- = separator;
		}
		break;
	}
	return dst;
}
//---------------------------------------------------------------------------
AnsiString __fastcall EnsureSign(AnsiString AStr)
{
	AnsiString str = AStr;
	switch(*str.c_str()) {
		case '-':
		case '+':
			break;
		default:
			str = "+" + str;
			break;
	}
	return str;
}
//----------------------------------------------------------------------------
void __fastcall AnsiTokenizeStr(TStrings* ATokens, AnsiString AStr,
	char ADelimiter)
{
	const char* start = AStr.c_str();
	const char* pos = start;
	ATokens->Clear();
	for(;;) {
		if(*pos == ADelimiter || *pos == '\0') {
			ATokens->Add(AnsiString(start, pos - start));
			if(*pos == '\0') {
				break;
			}
			start = pos = CharNext(pos);
			continue;
		}
		pos = CharNext(pos);
	}
}
//---------------------------------------------------------------------------
// variable type utilities
//---------------------------------------------------------------------------
DWORDLONG __fastcall MakeDwordLong(DWORD ALow, DWORD AHigh)
{
	return static_cast<DWORDLONG>(ALow) |
		(static_cast<DWORDLONG>(AHigh) << BITSPERDWORD);
}
//---------------------------------------------------------------------------
void __fastcall MakeDwords(DWORDLONG ADwordLong, DWORD& ALow, DWORD& AHigh)
{
	ALow = ADwordLong & FULLDWORD;
	AHigh = (ADwordLong >> BITSPERDWORD) & FULLDWORD;
}
//---------------------------------------------------------------------------
int __fastcall WrapInc(int AValue, int AMinValue, int AMaxValue)
{
	return AValue >= AMaxValue ? AMinValue : (AValue + 1);
}
//---------------------------------------------------------------------------
int __fastcall WrapDec(int AValue, int AMinValue, int AMaxValue)
{
	return AValue <= AMinValue ? AMaxValue : (AValue - 1);
}
//---------------------------------------------------------------------------
void __fastcall FlipBool(bool& AValue)
{
	AValue = !AValue;
}
//---------------------------------------------------------------------------

