//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TCompareDialog.h"
#include "TCompareLoader.h"
#include "TMessageDialog.h"
#include "MySysUtils.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TCompareDialog *CompareDialog;
//---------------------------------------------------------------------------
// __published:
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::FormShow(TObject* ASender)
{
	ReflectParams();
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::FormActivate(TObject* ASender)
{
	if(FFileName.Length() > 0) {
		TStringList* files = new TStringList();
		files->Add(FFileName);
		FilesDrop(files);
		delete files;
	}
	ActiveControl = NULL;
	SelectFirst();
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::FormCloseQuery(TObject* ASender,
	bool& ACanClose)
{
	bool canClose = true;
	if(ModalResult == Button1->ModalResult) {
		if(!UpdateParams()) {
			canClose = false;
		}
	}
	ACanClose = canClose;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::CheckBoxClick(TObject* ASender)
{
	UpdateEnabled();
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button11Click(TObject* ASender)
{
	SelectDir(SELECTFILE_SRC_DIR);
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button12Click(TObject* ASender)
{
	SelectDir(SELECTFILE_DST_DIR);
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button3Click(TObject* ASender)
{
	AnsiString fileName;
	if(DoSelectFile(fileName, SELECTFILE_IMPORT)) {
		ImportParams(fileName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button4Click(TObject* ASender)
{
	AnsiString fileName = "Untitled.sbc";
	if(UpdateParams() && DoSelectFile(fileName, SELECTFILE_EXPORT)) {
		ExportParams(fileName);
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button5Click(TObject* ASender)
{
	AnsiString temp = Edit11->Text;
	Edit11->Text = Edit12->Text;
	Edit12->Text = temp;
	ActivateControl(ActiveControl);
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::Button6Click(TObject* ASender)
{
	Options = TCompareOptions();
	OptionsChange();
	ActivateControl(ActiveControl);
}
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::SelectDir(int AId)
{
	for(;;) {
		TEdit* edit = NULL;
		switch(AId) {
			case SELECTFILE_SRC_DIR:
				edit = Edit11;
				break;
			case SELECTFILE_DST_DIR:
				edit = Edit12;
				break;
		}
		if(edit == NULL) {
			break;
		}
		AnsiString dir = edit->Text;
		if(!DoSelectFile(dir, AId)) {
			break;
		}
		edit->Text = dir;
		ActivateControl(edit);
		break;
	}
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::ReflectParams()
{
	ReflectOptions();
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::UpdateParams()
{
	return UpdateOptions();
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::ImportParams(AnsiString AFileName)
{
	bool succeeded = false;
	TCompareLoader* loader = new TCompareLoader();
	for(;;) {
		try {
			loader->LoadFromFile(AFileName);
		}
		catch(...) {
			Application->MessageBox(
				("The program cannot load data from the file: " +
					AFileName).c_str(),
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
			break;
		}
		if(!loader->IsSupported()) {
			if(Application->MessageBox(
				("The progream does not support the file: " + AFileName +
				"\r\n\r\nContinue?").c_str(),
				Application->Title.c_str(),
				MB_OKCANCEL | MB_ICONERROR | MB_SETFOREGROUND) != IDOK) {
				break;
			}
		}
		Options = loader->Options;
		DoFileAccess(AFileName, FILEACCESS_IMPORT);
		succeeded = true;
		break;
	}
	delete loader;
	return succeeded;
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::ExportParams(AnsiString AFileName)
{
	bool succeeded = false;
	TCompareLoader* loader = new TCompareLoader();
	for(;;) {
		try {
			loader->Options = Options;
			loader->SaveToFile(AFileName);
		}
		catch(...) {
			Application->MessageBox(
				("The program cannot save data to the file: " +
					AFileName).c_str(),
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
			break;
		}
		DoFileAccess(AFileName, FILEACCESS_EXPORT);
		succeeded = true;
		break;
	}
	delete loader;
	return succeeded;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::BackupParams()
{
	FBackup.Options = Options;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::RestoreParams()
{
	Options = FBackup.Options;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::UpdateEnabled()
{
	Edit13->Enabled = CheckBox11->Checked;
	Edit14->Enabled = CheckBox12->Checked;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::ActivateControl(TWinControl* AControl)
{
	if(AControl != NULL && AControl->CanFocus()) {
		ActiveControl = NULL;
		if(AControl->InheritsFrom(__classid(TEdit))) {
			dynamic_cast<TEdit*>(AControl)->SelectAll(); // necessary
		}
		ActiveControl = AControl;
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::SetOptions(TCompareOptions AOptions)
{
	if(AOptions != FOptions) {
		FOptions = AOptions;
		OptionsChange();
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::OptionsChange()
{
	ReflectOptions();
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::ReflectOptions()
{
	Edit11->Text = FOptions.SrcDir;
	Edit12->Text = FOptions.DstDir;
	CheckBox11->Checked = FOptions.UseIncludeMasks;
	Edit13->Text = FOptions.IncludeMasks;
	CheckBox12->Checked = FOptions.UseExcludeMasks;
	Edit14->Text = FOptions.ExcludeMasks;
	CheckBox13->Checked = FOptions.IncludeSubDirs;
	CheckBox21->Checked = FOptions.FatPrecision;
	CheckBox31->Checked = FOptions.ShowNotModified;
	UpdateEnabled();
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::UpdateOptions()
{
	bool succeeded = false;
	AnsiString errorMessage;
	TWinControl* errorControl = NULL;
	for(;;) {
		TCompareOptions options = Options;
		options.SrcDir = Edit11->Text;
		if(!DirectoryExists(options.SrcDir)) {
			errorMessage = "Set an exinsting directory.";
			errorControl = Edit11;
			break;
		}
		options.DstDir = Edit12->Text;
		if(!DirectoryExists(options.DstDir)) {
			errorMessage = "Set an exinsting directory.";
			errorControl = Edit12;
			break;
		}
		if(SameFileName(options.SrcDir, options.DstDir)) {
			errorMessage = "Set different directories.";
			errorControl = Edit11;
			break;
		}
		options.UseIncludeMasks = CheckBox11->Checked;
		options.IncludeMasks = Edit13->Text;
		if(options.UseIncludeMasks && options.IncludeMasks.Length() <= 0) {
			errorMessage = "Set masks.";
			errorControl = Edit13;
			break;
		}
		options.UseExcludeMasks = CheckBox12->Checked;
		options.ExcludeMasks = Edit14->Text;
		if(options.UseExcludeMasks && options.ExcludeMasks.Length() <= 0) {
			errorMessage = "Set masks.";
			errorControl = Edit14;
			break;
		}
		options.IncludeSubDirs = CheckBox13->Checked;
		options.FatPrecision = CheckBox21->Checked;
		options.ShowNotModified = CheckBox31->Checked;
		Options = options;
		succeeded = true;
		break;
	}
	if(!succeeded) {
		if(errorMessage.Length() > 0) {
			Application->MessageBox(
				errorMessage.c_str(),
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
		}
		if(errorControl != NULL) {
			ActivateControl(errorControl);
		}
	}
	return succeeded;
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::DoSelectFile(AnsiString& AFileName, int AId)
{
	bool selected = false;
	if(OnSelectFile != NULL) {
		OnSelectFile(this, AFileName, AId, selected);
	}
	return selected;
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::DoFileAccess(AnsiString AFileName, int AId)
{
	if(OnFileAccess != NULL) {
		OnFileAccess(this, AFileName, AId);
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::FilesDrop(const TStrings* AFiles)
{
	for(;;) {
		TStrings* files = const_cast<TStrings*>(AFiles);
		if(files->Count != 1) {
			Application->MessageBox(
				"Set one item.",
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
			break;
		}
		AnsiString fileName = files->Strings[0];
		if(FileExists(fileName)) {
			if(SameFileName(ExtractFileExt(fileName), ".lnk")) {
				fileName = ExtractFileLink(fileName);
			}
			else {
				ImportParams(fileName);
				break;
			}
		}
		TEdit* edit = NULL;
		{
			TMessageDialog* dialog = new TMessageDialog();
			dialog->Caption = Application->Title;
			dialog->Message = "Set the item to which?";   
			dialog->ButtonCaptions->Add("&Source");
			dialog->ButtonCaptions->Add("&Destination");
			dialog->ButtonCaptions->Add("Cancel");
			dialog->IconType = TMessageDialog::ICONTYPE_WARNING;
			dialog->DefaultButton = 0;
			dialog->CancelButton = 2;
			switch(dialog->Execute()) {
				case 0: edit = Edit11; break;
				case 1: edit = Edit12; break;
			}
			delete dialog;
		}
		if(edit == NULL) {
			break;
		}
		edit->Text = fileName;
		ActivateControl(edit);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareDialog::WMDropFiles(TWMDropFiles& AMessage)
{
	SetForegroundWindow(Application->Handle);
	TStringList* files = new TStringList();
	{
		HDROP handle = reinterpret_cast<HDROP>(AMessage.Drop);
		int fileCount = DragQueryFile(handle, 0xFFFFFFFF, NULL, 0);
		const int bufferSize = MAX_PATH;
		char buffer[bufferSize];
		for(int i = 0; i < fileCount; i++) {
			if(DragQueryFile(handle, i, buffer, bufferSize) > 0) {
				files->Add(ShortFileNameToLongFileName(buffer));
			}
		}
		DragFinish(handle);
	}
	if(files->Count > 0) {
		FilesDrop(files);
	}
	delete files;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TCompareDialog::TCompareDialog(TComponent* AOwner) :
	TForm(AOwner),
	FFileName(),
	FOptions(),
	FBackup(),
	FOnSelectFile(NULL),
	FOnFileAccess(NULL)
{
	DragAcceptFiles(Handle, true);
	OptionsChange();
}
//---------------------------------------------------------------------------
__fastcall TCompareDialog::~TCompareDialog()
{
	FOnFileAccess = NULL;
	FOnSelectFile = NULL;
}
//---------------------------------------------------------------------------
bool __fastcall TCompareDialog::Execute(AnsiString AFileName)
{
	bool ok = false;
	{
		FFileName = AFileName;
		BackupParams();
	}
	if(ShowModal() == Button1->ModalResult) {
		ok = true;
	}
	{
		if(!ok) {
			RestoreParams();
		}
		FFileName = "";
	}
	return ok;
}
//---------------------------------------------------------------------------

