//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TCompareThread.h"
#include "MySysUtils.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// internal
//---------------------------------------------------------------------------
namespace
{
const __int64 FAT_PRECISION = 2 * 1000 * 1000 * 10;
} // end of namespace
//---------------------------------------------------------------------------
// external
//---------------------------------------------------------------------------
// class TCompareThread : public TReportThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::IsFileFound(HANDLE AHandle)
{
	return AHandle != INVALID_HANDLE_VALUE && AHandle != NULL;
}
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::IsRelativeDir(const WIN32_FIND_DATA& AData)
{
	return
		lstrcmp(AData.cFileName, ".") == 0 ||
		lstrcmp(AData.cFileName, "..") == 0;
}
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::IsDirectory(const WIN32_FIND_DATA& AData)
{
	return (AData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
}
//---------------------------------------------------------------------------
__int64 __fastcall TCompareThread::MakeSize(const WIN32_FIND_DATA& AData)
{
	return static_cast<__int64>(MakeDwordLong(AData.nFileSizeLow,
		AData.nFileSizeHigh));
}
//---------------------------------------------------------------------------
__int64 __fastcall TCompareThread::MakeSizeRem(__int64 AValue1,
	__int64 AValue2)
{
	return AValue1 - AValue2;
}
//---------------------------------------------------------------------------
__int64 __fastcall TCompareThread::MakeTime(const WIN32_FIND_DATA& AData)
{
	return static_cast<__int64>(MakeDwordLong(
		AData.ftLastWriteTime.dwLowDateTime,
		AData.ftLastWriteTime.dwHighDateTime));
}
//---------------------------------------------------------------------------
__int64 __fastcall TCompareThread::MakeTimeRem(__int64 AValue1,
	__int64 AValue2, bool AFatPrecision)
{
	return AFatPrecision ?
		(MakeFatTime(AValue1) - MakeFatTime(AValue2)) :
		(AValue1 - AValue2);
}
//---------------------------------------------------------------------------
__int64 __fastcall TCompareThread::MakeFatTime(__int64 AValue)
{
	return ((AValue + (FAT_PRECISION - 1)) / FAT_PRECISION) * FAT_PRECISION;
}
//---------------------------------------------------------------------------
int __fastcall TCompareThread::MakeResult(__int64 ASizeRem, __int64 ATimeRem)
{
	int result;
	for(;;) {
		if(ATimeRem > 0) {
			result = TCompareItem::RESULT_MODIFIED;
			break;
		}
		if(ASizeRem == 0) {
			result = TCompareItem::RESULT_NOT_MODIFIED;
			break;
		}
		result = TCompareItem::RESULT_AMBIGUOUS;
		break;
	}
	return result;
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::StrToMasks(AnsiString AStr, TStrings* AMasks)
{
	AnsiTokenizeStr(AMasks, AStr, ';');
}
//---------------------------------------------------------------------------
int __fastcall TCompareThread::CompareItem(void* AItem1, void* AItem2)
{
	TCompareItem* item1 = static_cast<TCompareItem*>(AItem1);
	TCompareItem* item2 = static_cast<TCompareItem*>(AItem2);
	int result = item1->Result - item2->Result;
	if(result == 0) {
		result = AnsiCompareFileName(item1->SrcName, item2->SrcName);
	}
	return result;
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::PrepareMasks()
{
	if(FOptions.UseIncludeMasks) {
		StrToMasks(FOptions.IncludeMasks, FIncludeMasks);
	}
	if(FOptions.UseExcludeMasks) {
		StrToMasks(FOptions.ExcludeMasks, FExcludeMasks);
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::EnumFiles()
{
	EnumFiles("");
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::EnumFiles(AnsiString ADir)
{
	AnsiString path;
	if(ADir.Length() > 0) {
		path = IncludeTrailingPathDelimiter(ADir);
	}
	AnsiString srcPath = IncludeTrailingPathDelimiter(FOptions.SrcDir) + path;
	AnsiString dstPath = IncludeTrailingPathDelimiter(FOptions.DstDir) + path;
	WIN32_FIND_DATA data;
	HANDLE handle = FindFirstFile((srcPath + "*").c_str(), &data);
	if(IsFileFound(handle)) {
		AnsiString name;
		TCompareItem item;
		do {
			if(Terminated) {
				break;
			}
			for(;;) {
				if(IsRelativeDir(data)) {
					break;
				}
				name = data.cFileName;
				if(IsDirectory(data)) {
					if(FOptions.IncludeSubDirs) {
						EnumFiles(path + name);
					}
					break;
				}
				if(!FilterByName(name)) {
					break;
				}
				item.SrcName = srcPath + name;
				item.DstName = dstPath + name;
				item.SizeRem = MakeSize(data);
				item.TimeRem = MakeTime(data);
				FItems->Add(new TCompareItem(item));
				break;
			}
		} while(FindNextFile(handle, &data));
		FindClose(handle);
	}
}
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::FilterByName(AnsiString AFileName) const
{
	return
		(!FOptions.UseIncludeMasks || MatchByName(AFileName, FIncludeMasks)) &&
		(!FOptions.UseExcludeMasks || !MatchByName(AFileName, FExcludeMasks));
}
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::MatchByName(AnsiString AFileName,
	const TStrings* AMasks) const
{
	bool matched = false;
	TStrings* masks = const_cast<TStrings*>(AMasks);
	for(int i = 0; i < masks->Count; i++) {
		if(Terminated) {
			break;
		}
		if(MatchFileName(AFileName, masks->Strings[i])) {
			matched = true;
			break;
		}
	}
	return matched;
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::CompareFiles()
{
	int processed = 0;
	int all = FItems->Count;
	TCompareItem* item;
	for(int i = 0; i < FItems->Count; i++) {
		if(Terminated) {
			break;
		}
		item = static_cast<TCompareItem*>(FItems->Items[i]);
		CompareFile(*item);
		if(!FilterByResult(item->Result)) {
			delete item; FItems->Delete(i--);
		}
		DoProgress(++processed, all);
	}
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::CompareFile(TCompareItem& AItem)
{
	for(;;) {
		WIN32_FIND_DATA data;
		HANDLE handle = FindFirstFile(AItem.DstName.c_str(), &data);
		if(!IsFileFound(handle)) {
			AItem.Duplicates = false;
			AItem.TimeRem = 0;
			AItem.Result = TCompareItem::RESULT_CREATED;
			break;
		}
		FindClose(handle);
		AItem.Duplicates = true;
		if(IsDirectory(data) || AItem.SizeRem < 0 || MakeSize(data) < 0 ||
			AItem.TimeRem <= 0 || MakeTime(data) <= 0) {
			AItem.SizeRem = 0;
			AItem.TimeRem = 0;
			AItem.Result = TCompareItem::RESULT_ERROR;
			break;
		}
		AItem.SizeRem = MakeSizeRem(AItem.SizeRem, MakeSize(data));
		AItem.TimeRem = MakeTimeRem(AItem.TimeRem, MakeTime(data),
			FOptions.FatPrecision);
		AItem.Result = MakeResult(AItem.SizeRem, AItem.TimeRem);
		break;
	}
}
//---------------------------------------------------------------------------
bool __fastcall TCompareThread::FilterByResult(int AResult) const
{
	return
		FOptions.ShowNotModified ||
		AResult != TCompareItem::RESULT_NOT_MODIFIED;
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::SortItems()
{
	FItems->Sort(CompareItem);
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TCompareThread::Prepare()
{
	PrepareMasks();
	EnumFiles();
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::Process()
{
	CompareFiles();
	SortItems();
}
//---------------------------------------------------------------------------
void __fastcall TCompareThread::Unprepare()
{
	delete FExcludeMasks; FExcludeMasks = NULL;
	delete FIncludeMasks; FIncludeMasks = NULL;
	for(int i = 0; i < FItems->Count; i++) {
		delete static_cast<TCompareItem*>(FItems->Items[i]);
	}
	delete FItems; FItems = NULL;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TCompareThread::TCompareThread(TCompareOptions AOptions) :
	TReportThread(true),
	FOptions(AOptions),
	FItems(new TList()),
	FIncludeMasks(new TStringList()),
	FExcludeMasks(new TStringList())
{
	FreeOnTerminate = true;
}
//---------------------------------------------------------------------------

