//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <FileCtrl.hpp>
#include <Clipbrd.hpp>
#include <InIFiles.hpp>
#include "TMainForm.h"
#include "MySysUtils.h"
#include "TCompareDialog.h"
#include "TProgressDialog.h"
#include "TCompareThread.h"
#include "TCompareItem.h"
#include "TCompareOptions.h"
#include "TOperationsDialog.h"
#include "TFindItemDialog.h"
#include "TOpenDialogList.h"
#include "TFileHistory.h"
#include "TMessageDialog.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TMainForm* MainForm = NULL;
//---------------------------------------------------------------------------
// internal
//---------------------------------------------------------------------------
namespace
{
const AnsiString ABOUT_VERSION = "Version 1.0.0.";
const AnsiString FILE_PATH = ExtractFilePath(Application->ExeName);
const AnsiString FILE_HELP = FILE_PATH + "Help\\index.html";
const AnsiString FILE_INI = FILE_PATH + "Settings.ini";
const AnsiString INI_VIEWOPTIONS = "ViewOptions";
const AnsiString INI_VIEWOPTIONS_TOOLBAR = "ToolBar";
const AnsiString INI_VIEWOPTIONS_STATUSBAR = "StatusBar";
const AnsiString INI_VIEWOPTIONS_LASTPOSITION = "LastPosition";
const AnsiString INI_VIEWOPTIONS_TOPMOST = "TopMost";
const AnsiString INI_MAINFORM = "MainForm";
const AnsiString INI_MAINFORM_WINDOWSTATE = "WindowState";
const AnsiString INI_MAINFORM_LEFT = "Left";
const AnsiString INI_MAINFORM_TOP = "Top";
const AnsiString INI_MAINFORM_RIGHT = "Right";
const AnsiString INI_MAINFORM_BOTTOM = "Bottom";
const AnsiString INI_MAINFORM_LISTCOLUMNWIDTH = "ListColumnWidth";
const AnsiString INI_FILEHISTORY = "FileHistory";
const AnsiString INI_FILEHISTORY_ITEM = "Item";
const AnsiString INI_OPENDIALOGS = "OpenDialogs";
const AnsiString INI_OPENDIALOGS_INITIALDIR = "InitialDir";
const AnsiString INI_OPENDIALOGS_FILTERINDEX = "FilterIndex";
} // end of namespace
//---------------------------------------------------------------------------
// external
//---------------------------------------------------------------------------
// class TMainForm : public TForm
//---------------------------------------------------------------------------
// __published:
//---------------------------------------------------------------------------
void __fastcall TMainForm::ApplicationIdle(TObject* ASender, bool& ADone)
{
	ADone = true;
	Application->OnIdle = NULL;
	ProcessCommandParams();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FormCloseQuery(TObject* ASender, bool& ACanClose)
{
	ACanClose = ClearList(true);
	if(ACanClose) {
		SaveSettings();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareM1Click(TObject* ASender)
{
	CompareFiles("");
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SaveListAsM1Click(TObject* ASender)
{
	for(;;) {
		AnsiString fileName = "Untitled.csv";
		if(!OpenDialogs->SelectFiles(SELECTFILES_SAVE_LIST_AS, fileName,
			NULL)) {
			break;
		}
		try {
			SaveListItemsToFile(fileName, ListView1, true);
		}
		catch(...) {
			Application->MessageBox(
				("The program cannot save data to the file: " +
				fileName).c_str(),
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
		}
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::HistoryM1Click(TObject* ASender)
{
	for(;;) {
		int index = FileHistoryMenuItems->IndexOf(ASender);
		if(index < 0) {
			break;
		}
		AnsiString fileName = FileHistory->Items[index];
		if(!FileExists(fileName)) {
			if(Application->MessageBox(
				("The program cannot find the file: " + fileName +
				"\r\n\r\nDelete the item from the history?").c_str(),
				Application->Title.c_str(),
				MB_OKCANCEL | MB_ICONERROR | MB_SETFOREGROUND) == IDOK) {
				FileHistory->Delete(index);
			}
			break;
		}
		CompareFiles(fileName);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ClearM1Click(TObject* ASender)
{
	if(Application->MessageBox(
		"Clear the history?",
		Application->Title.c_str(),
		MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) == IDOK) {
		FileHistory->Clear();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ExitM1Click(TObject* ASender)
{
	Close();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CopyM1Click(TObject* ASender)
{
	TStringList* lines = new TStringList();
	AnsiString line;
	TCompareItem* item;
	for(int i = 0; i < ListView1->Items->Count; i++) {
		if(ListView1->Items->Item[i]->Selected) {
			item = static_cast<TCompareItem*>(ListView1->Items->Item[i]->Data);
			if(ASender == SourceM1) {
				line = item->SrcName;
			}
			else if(ASender == DestinationM1) {
				line = item->DstName;
			}
			else {
				break;
			}
			lines->Add(line);
		}
	}
	Clipboard()->AsText = lines->Text;
	delete lines;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::DeleteM1Click(TObject* ASender)
{
	for(;;) {
		if(Application->MessageBox(
			"Delete the selected items?",
			Application->Title.c_str(),
			MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
			break;
		}
		ListView1->DeleteSelected();
		ItemCountChange();
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SelectM1Click(TObject* ASender)
{
	try {
		ListView1->Items->BeginUpdate();
		if(ASender == AllM1) {
			ListView1->SelectAll();
		}
		else if(ASender == NoneM1) {
			ListView1->ClearSelection();
		}
		else if(ASender == InvertM1) {
			for(int i = 0; i < ListView1->Items->Count; i++) {
				ListView1->Items->Item[i]->Selected =
					!ListView1->Items->Item[i]->Selected;
			}
		}
	}
	__finally {
		ListView1->Items->EndUpdate();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SearchM1Click(TObject* ASender)
{
	for(;;) {
		bool select = false;
		if(ASender == FindM1) {
			select = true;
		}
		else if(ASender == FindNextM1 || ASender == FindPreviousM1) {
			if(FindItemDialog->Options.FindText.Length() <= 0) {
				select = true;
			}
		}
		else {
			break;
		}
		if(select) {
			if(!FindItemDialog->Execute()) {
				break;
			}
		}
		TFindItemOptions options = FindItemDialog->Options;
		if(ASender == FindPreviousM1) {
			options.Down = !options.Down;
		}
		FindItem(options);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SortAscendM1Click(TObject* ASender)
{
	int flags = 0;
	if(ASender == ByNameM1) {
		flags |= SORT_BY_NAME;
	}
	else if(ASender == ByDirectoryM1) {
		flags |= SORT_BY_DIRECTORY;
	}
	else if(ASender == ByExtensionM1) {
		flags |= SORT_BY_EXTENSION;
	}
	else if(ASender == BySizeRemainderM1) {
		flags |= SORT_BY_SIZE_REMAINDER;
	}
	else if(ASender == ByTimeRemainderM1) {
		flags |= SORT_BY_TIME_REMAINDER;
	}
	else if(ASender == ByResultM1) {
		flags |= SORT_BY_RESULT;
	}
	else if(ASender == ByAbsoluteNameM1) {
		flags |= SORT_BY_ABSOLUTE_NAME;
	}
	SortItems(flags);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SortDescendM1Click(TObject* ASender)
{
	int flags = 0;
	if(ASender == ByNameM2) {
		flags |= SORT_BY_NAME | SORT_DESCEND;
	}
	else if(ASender == ByDirectoryM2) {
		flags |= SORT_BY_DIRECTORY | SORT_DESCEND;
	}
	else if(ASender == ByExtensionM2) {
		flags |= SORT_BY_EXTENSION | SORT_DESCEND;
	}
	else if(ASender == BySizeRemainderM2) {
		flags |= SORT_BY_SIZE_REMAINDER | SORT_DESCEND;
	}
	else if(ASender == ByTimeRemainderM2) {
		flags |= SORT_BY_TIME_REMAINDER | SORT_DESCEND;
	}
	else if(ASender == ByResultM2) {
		flags |= SORT_BY_RESULT | SORT_DESCEND;
	}
	else if(ASender == ByAbsoluteNameM2) {
		flags |= SORT_BY_ABSOLUTE_NAME | SORT_DESCEND;
	}
	SortItems(flags);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SizeRemaindersM1Click(TObject* ASender)
{
	int selected = 0;
	__int64 sum = 0;
	for(int i = 0; i < ListView1->Items->Count; i++) {
		if(ListView1->Items->Item[i]->Selected) {
			selected++;
			sum += static_cast<TCompareItem*>(
				ListView1->Items->Item[i]->Data)->SizeRem;
		}
	}
	AnsiString message = AnsiString(
		"Selected: " + FormatInt(selected) + " items\r\n"
		"Sum: " + FormatInt(sum) + " bytes");
	Application->MessageBox(
		message.c_str(),
		"Size Remainders",
		MB_OK | MB_ICONINFORMATION | MB_SETFOREGROUND);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ExploreM1Click(TObject* ASender)
{
	AccessItems(
		ASender == SourceM2 || ASender == BothM1,
		ASender == DestinationM2 || ASender == BothM1,
		ACCESSITEMS_EXPLORE);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PropertiesM1Click(TObject* ASender)
{
	AccessItems(
		ASender == SourceM3 || ASender == BothM2,
		ASender == DestinationM3 || ASender == BothM2,
		ACCESSITEMS_PROPERTIES);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ViewM1Click(TObject* ASender)
{
	for(;;) {
		TViewOptions viewOptions = ViewOptions;
		if(ASender == ToolbarM1) {
			FlipBool(viewOptions.ToolBar);
		}
		else if(ASender == StatusBarM1) {
			FlipBool(viewOptions.StatusBar);
		}
		else if(ASender == LastPositionM1) {
			FlipBool(viewOptions.LastPosition);
		}
		else if(ASender == TopmostM1) {
			FlipBool(viewOptions.Topmost);
		}
		else if(ASender == DefaultM1) {
			if(Application->MessageBox(
				"Set default settings?",
				Application->Title.c_str(),
				MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
				break;
			}
			viewOptions = TViewOptions();
		}
		else {
			break;
		}
		ViewOptions = viewOptions;
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ResizeListColumnsM1Click(TObject* ASender)
{
	try {
		ListView1->Columns->BeginUpdate();
		int width = ListView1->ClientWidth / ListView1->Columns->Count;
		for(int i = 0; i < ListView1->Columns->Count; i++) {
			ListView1->Columns->Items[i]->Width = 0;
			ListView1->Columns->Items[i]->Width = width;
		}
	}
	__finally {
		ListView1->Columns->EndUpdate();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::OperationsM1Click(TObject* ASender)
{
	TStringList* from = new TStringList();
	TStringList* to = new TStringList();
	bool selectionOnly = ASender == PerformToSelectionM1;
	TCompareItem* item;
	for(int i = 0; i < ListView1->Items->Count; i++) {
		if(!selectionOnly || ListView1->Items->Item[i]->Selected) {
			item = static_cast<TCompareItem*>(ListView1->Items->Item[i]->Data);
			from->Add(item->SrcName);
			to->Add(item->DstName);
		}
	}
	OperationsDialog->Execute(from, to);
	delete to;
	delete from;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::HelpM2Click(TObject* ASender)
{
	for(;;) {
		if(Application->MessageBox(
			"Open the help document?",
			Application->Title.c_str(),
			MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
			break;
		}
		bool succeeded = reinterpret_cast<int>(ShellExecute(
			Application->Handle, "open", FILE_HELP.c_str(), NULL, NULL,
			SW_SHOWNORMAL)) >= 32;
		if(!succeeded) {
			Application->MessageBox(
				("The program cannot open the document: " + FILE_HELP).c_str(),
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
			break;
		}
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::AboutM1Click(TObject* ASender)
{
	TMessageDialog* dialog = new TMessageDialog();
	dialog->Caption = "About";
	dialog->Message =
		Application->Title + " for Windows " + ABOUT_VERSION + "\r\n\r\n" +
		"Copyright (C) 2005 Moochos. All rights reserved.";
	dialog->ButtonCaptions->Add("OK");
	dialog->IconType = TMessageDialog::ICONTYPE_OTHER;
	dialog->Icon->Assign(Application->Icon);
	dialog->DefaultButton = 0;
	dialog->CancelButton = 0;
	dialog->Execute();
	delete dialog;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::PopupMenu1Click(TObject* ASender)
{
	if(ASender == SourceP1) {
		SourceM1->Click();
	}
	else if(ASender == DestinationP1) {
		DestinationM1->Click();
	}
	else if(ASender == DeleteP1) {
		DeleteM1->Click();
	}
	else if(ASender == AllP1) {
		AllM1->Click();
	}
	else if(ASender == NoneP1) {
		NoneM1->Click();
	}
	else if(ASender == InvertP1) {
		InvertM1->Click();
	}
	else if(ASender == FindP1) {  
		FindM1->Click();
	}
	else if(ASender == FindNextP1) {
		FindNextM1->Click();
	}
	else if(ASender == FindPreviousP1) {
		FindPreviousM1->Click();
	}
	else if(ASender == SourceP2) {
		SourceM2->Click();
	}
	else if(ASender == DestinationP2) {
		DestinationM2->Click();
	}
	else if(ASender == BothP1) {
		BothM1->Click();
	}
	else if(ASender == SourceP3) {
		SourceM3->Click();
	}
	else if(ASender == DestinationP3) {
		DestinationM3->Click();
	}
	else if(ASender == BothP2) {
		BothM2->Click();
	}
	else if(ASender == SizeRemaindersP1) {
		SizeRemaindersM1->Click();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ToolButtonClick(TObject* ASender)
{
	if(ASender == CompareT1) {
		CompareM1->Click();
	}
	else if(ASender == CopySourceT1) {
		SourceM1->Click();
	}
	else if(ASender == CopyDestinationT1) {
		DestinationM1->Click();
	}
	else if(ASender == DeleteItemsT1) {
		DeleteM1->Click();
	}
	else if(ASender == SelectAllT1) {
		AllM1->Click();
	}
	else if(ASender == FindT1) {
		FindM1->Click();
	}
	else if(ASender == PerformToSelectionT1) {
		PerformToSelectionM1->Click();
	}
	else if(ASender == PerformToAllT1) {
		PerformToAllM1->Click();
	}
	else if(ASender == ResizeListColumnsT1) {
		ResizeListColumnsM1->Click();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ListView1Deletion(TObject* ASender,
	TListItem* AItem)
{
	delete static_cast<TCompareItem*>(AItem->Data);
	AItem->Data = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ListView1SelectItem(TObject* ASender,
	TListItem* AItem, bool ASelected)
{
	UpdateEnabled(false);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ListView1ColumnClick(TObject* ASender,
	TListColumn* AColumn)
{
	switch(AColumn->Index) {
		case 0: ByNameM1->Click(); break;
		case 1: ByDirectoryM1->Click(); break;
		case 2: ByExtensionM1->Click(); break;
		case 3: BySizeRemainderM1->Click(); break;
		case 4: ByTimeRemainderM1->Click(); break;
		case 5: ByResultM1->Click(); break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareDialogSelectFile(TObject* ASender,
	AnsiString& AFileName, int AId, bool& ASelected)
{
	switch(AId) {
		case TCompareDialog::SELECTFILE_SRC_DIR:
		case TCompareDialog::SELECTFILE_DST_DIR:
			ASelected = OpenDialogs->SelectDirectory("Select a directory.",
				AFileName);
			break;
		case TCompareDialog::SELECTFILE_IMPORT:
			ASelected = OpenDialogs->SelectFiles(SELECTFILES_COMPARE_IMPORT,
				AFileName, NULL);
			break;
		case TCompareDialog::SELECTFILE_EXPORT:
			ASelected = OpenDialogs->SelectFiles(SELECTFILES_COMPARE_EXPORT,
				AFileName, NULL);
			break;
		default:
			ASelected = false;
			break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareDialogFileAccess(TObject* ASender,
	AnsiString AFileName, int AId)
{
	FileHistory->Add(AFileName);
	OpenDialogs->InitialDir = ExtractFileDir(AFileName);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareThreadProgress(TObject* ASender,
	int AProcessed, int AAll)
{
	ProgressDialog->ShowProgress(AProcessed, AAll);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareThreadTerminate(TObject* ASender)
{
	for(;;) {
		TCompareThread* thread = dynamic_cast<TCompareThread*>(ASender);
		if(thread->Error || !thread->IsFinished()) {
			ProgressDialog->Finish(false);
			ProgressDialog->ShowMessage(thread->Error ? "Error" : "Canceled");
			break;
		}
		if(!ClearList(true)) {
			ProgressDialog->Finish(true);
			break;
		}
		try {
			ListView1->Items->BeginUpdate();
			TListItem* listItem;
			for(int i = 0; i < thread->Items->Count; i++) {
				listItem = ListView1->Items->Add();
				listItem->Data = new TCompareItem(*static_cast<TCompareItem*>(
					thread->Items->Items[i]));
				UpdateItem(listItem->Index);
			}
			if(ListView1->Items->Count > 0) {
				ListView1->ClearSelection();
				ListView1->ItemFocused = ListView1->Items->Item[0];
			}
			ItemCountChange();
		}
		__finally {
			ListView1->Items->EndUpdate();
		}
		ProgressDialog->Finish(true);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FileHistoryChange(TObject* ASender)
{
	TMenuItem* menuItem;
	AnsiString fileName;
	for(int i = 0; i < FileHistory->Capacity; i++) {
		menuItem = static_cast<TMenuItem*>(FileHistoryMenuItems->Items[i]);
		if(i < FileHistory->Count) {
			menuItem->Visible = true;
			fileName = FileHistory->Items[i];
			menuItem->Caption = AnsiString("&") + (i + 1) + " " +
				ExtractFileName(fileName) + " - " + ExtractFileDir(fileName);
		}
		else {
			menuItem->Visible = false;
			menuItem->Caption = "";
		}
	}
	UpdateEnabled(false);
}
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
int CALLBACK TMainForm::CompareItem(LPARAM ALParam1, LPARAM ALParam2,
	LPARAM ALParam)
{
	int result = 0;
	TListItem* item1 = reinterpret_cast<TListItem*>(ALParam1);
	TListItem* item2 = reinterpret_cast<TListItem*>(ALParam2);
	if((ALParam & SORT_BY_NAME) != 0) {
		result = AnsiCompareFileName(
			item1->Caption,
			item2->Caption);
	}
	else if((ALParam & SORT_BY_DIRECTORY) != 0) {
		result = AnsiCompareFileName(
			item1->SubItems->Strings[0],
			item2->SubItems->Strings[0]);
	}
	else if((ALParam & SORT_BY_EXTENSION) != 0) {
		result = AnsiCompareFileName(
			item1->SubItems->Strings[1],
			item2->SubItems->Strings[1]);
	}
	else if((ALParam & SORT_BY_SIZE_REMAINDER) != 0) {
		result =
			static_cast<TCompareItem*>(item1->Data)->SizeRem -
			static_cast<TCompareItem*>(item2->Data)->SizeRem;
	}
	else if((ALParam & SORT_BY_TIME_REMAINDER) != 0) {
		if(static_cast<TCompareItem*>(item1->Data)->Duplicates &&
			static_cast<TCompareItem*>(item2->Data)->Duplicates) {
			result =
				static_cast<TCompareItem*>(item1->Data)->TimeRemInDays -
				static_cast<TCompareItem*>(item2->Data)->TimeRemInDays;
		}
		else if(static_cast<TCompareItem*>(item1->Data)->Duplicates) {
			result = -1;
		}
		else if(static_cast<TCompareItem*>(item2->Data)->Duplicates) {
			result = 1;
		}
	}
	else if((ALParam & SORT_BY_RESULT) != 0) {
		result =
			static_cast<TCompareItem*>(item1->Data)->Result -
			static_cast<TCompareItem*>(item2->Data)->Result;
	}
	else if((ALParam & SORT_BY_ABSOLUTE_NAME) != 0) {
		result = AnsiCompareFileName(
			static_cast<TCompareItem*>(item1->Data)->SrcName,
			static_cast<TCompareItem*>(item2->Data)->SrcName);
	}
	if(result == 0) {
		result = item1->Index - item2->Index;
	}
	if((ALParam & SORT_DESCEND) != 0) {
		result *= -1;
	}
	return result;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::CompareFiles(AnsiString AFileName)
{
	for(;;) {
		if(!CompareDialog->Execute(AFileName)) {
			break;
		}
		if(!ClearList(true)) {
			break;
		}
		TCompareThread* thread = new TCompareThread(CompareDialog->Options);
		thread->OnProgress = CompareThreadProgress;
		thread->OnTerminate = CompareThreadTerminate;
		ProgressDialog->Caption = "Compare";
		ProgressDialog->Start(thread);
		break;
	}
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::ClearList(bool AConfirm)
{
	bool succeeded = false;
	for(;;) {
		if(AConfirm && ListView1->Items->Count > 0) {
			if(Application->MessageBox(
				"Clear the list?",
				Application->Title.c_str(),
				MB_ICONEXCLAMATION | MB_OKCANCEL | MB_SETFOREGROUND) != IDOK) {
				break;
			}
		}
		ListView1->Clear();
		ItemCountChange();
		succeeded = true;
		break;
	}
	return succeeded;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FindItem(TFindItemOptions AOptions)
{
	if(ListView1->Items->Count > 0) {
		int start = 0;
		{
			TListItem* item = ListView1->ItemFocused;
			if(item != NULL) {
				start = item->Index;
			}
			if(ListView1->Items->Item[start]->Selected) {
				start = AOptions.Down ?
					WrapInc(start, 0, ListView1->Items->Count - 1) :
					WrapDec(start, 0, ListView1->Items->Count - 1);
			}
		}
		int current = start;
		AnsiString fileName;
		ListView1->ClearSelection();
		do {
			if(AOptions.MatchAbsoluteName) {
				fileName = static_cast<TCompareItem*>(
					ListView1->Items->Item[current]->Data)->SrcName;
			}
			else {
				fileName = ListView1->Items->Item[current]->Caption;
			}
			if(MatchFileName(fileName, AOptions.FindText)) {
				ListView1->Items->Item[current]->Selected = true;
				if(!AOptions.FindAll) {
					break;
				}
			}
			current = AOptions.Down ?
				WrapInc(current, 0, ListView1->Items->Count - 1) :
				WrapDec(current, 0, ListView1->Items->Count - 1);
		} while(current != start);
	}
	TListItem* selected = ListView1->Selected;
	if(selected != NULL) {
		ListView1->ItemFocused = selected;
		SendMessage(ListView1->Handle, LVM_ENSUREVISIBLE, selected->Index,
			false);
	}
	else {
		Application->MessageBox(
			"There are no matches.",
			Application->Title.c_str(),
			MB_OK | MB_ICONINFORMATION | MB_SETFOREGROUND);
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SortItems(int AFlags)
{
	TCursor cursor = Screen->Cursor; Screen->Cursor = crHourGlass;
	try {
		ListView1->CustomSort(CompareItem, AFlags);
	}
	__finally {
		Screen->Cursor = cursor;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::AccessItems(bool ASrc, bool ADst, int AOperation)
{
	TStringList* files = NULL;
	for(;;) {
		TListItem* selected = ListView1->Selected;
		if(selected == NULL) {
			break;
		}
		files = new TStringList();
		if(ASrc) {
			files->Add(static_cast<TCompareItem*>(selected->Data)->SrcName);
		}
		if(ADst) {
			files->Add(static_cast<TCompareItem*>(selected->Data)->DstName);
		}
		for(int i = 0; i < files->Count; i++) {
			if(!FileExists(files->Strings[i])) {
				Application->MessageBox(
					("The program cannot find the file: " +
					files->Strings[i]).c_str(),
					Application->Title.c_str(),
					MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
				continue;
			}
			switch(AOperation) {
				case ACCESSITEMS_EXPLORE:
					ShellExecute(Application->Handle, "explore",
						ExtractFileDir(files->Strings[i]).c_str(), NULL, NULL,
						SW_SHOWNORMAL);
					break;
				case ACCESSITEMS_PROPERTIES:
					ShowFileProperties(files->Strings[i]);
					break;
			}
		}
		break;
	}
	delete files;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::InitOpenDialogs()
{
	OpenDialogs->Count = SELECTFILES_COUNT;
	OpenDialogs->Items[SELECTFILES_SAVE_LIST_AS] = SaveDialog1;
	OpenDialogs->Items[SELECTFILES_COMPARE_IMPORT] = OpenDialog1;
	OpenDialogs->Items[SELECTFILES_COMPARE_EXPORT] = SaveDialog2;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UninitOpenDialogs()
{
	OpenDialogs->Clear();
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::InitFileHistory()
{
	TMenuItem* menuItems[] = {
		HistoryItemM1,
		HistoryItemM2,
		HistoryItemM3,
		HistoryItemM4,
		HistoryItemM5,
		HistoryItemM6,
		HistoryItemM7,
		HistoryItemM8,
	};
	int menuItemCount = sizeof(menuItems) / sizeof(TMenuItem*);
	for(int i = 0; i < menuItemCount; i++) {
		FileHistoryMenuItems->Add(menuItems[i]);
	}
	FileHistory->Capacity = menuItemCount;
	FileHistory->OnChange = FileHistoryChange;
	FileHistory->OnChange(FileHistory);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UninitFileHistory()
{
	// do nothing
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::InitCompareDialog()
{
	CompareDialog = new TCompareDialog(this);
	CompareDialog->OnSelectFile = CompareDialogSelectFile;
	CompareDialog->OnFileAccess = CompareDialogFileAccess;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UninitCompareDialog()
{
	delete CompareDialog; CompareDialog = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ProcessCommandParams()
{
	if(ParamCount() > 0) {
		TStringList* params = new TStringList();
		for(int i = 0; i < ParamCount(); i++) {
			params->Add(ShortFileNameToLongFileName(ParamStr(i + 1)));
		}
		FilesDrop(params);
		delete params;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::LoadSettings()
{
	for(;;) {
		if(!FileExists(FILE_INI)) {
			break;
		}
		if(TryLoadSettings()) {
			break;
		}
		if(Application->MessageBox(
			"The program cannot load the settings.\r\n\r\nRetry?",
			Application->Title.c_str(),
			MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
			break;
		}
	}
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::TryLoadSettings()
{
	bool succeeded = false;
	TMemIniFile* ini = NULL;
	try {
		ini = new TMemIniFile(FILE_INI);
		{
			TViewOptions viewOptions;
			viewOptions.ToolBar = ini->ReadBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_TOOLBAR,
				viewOptions.ToolBar);
			viewOptions.StatusBar = ini->ReadBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_STATUSBAR,
				viewOptions.StatusBar);
			viewOptions.LastPosition = ini->ReadBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_LASTPOSITION,
				viewOptions.LastPosition);
			viewOptions.Topmost = ini->ReadBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_TOPMOST,
				viewOptions.Topmost);
			ViewOptions = viewOptions;
		}
		if(ViewOptions.LastPosition) {
			WindowState = static_cast<TWindowState>(
				ini->ReadInteger(
				INI_MAINFORM,
				INI_MAINFORM_WINDOWSTATE,
				wsNormal));
			if(WindowState == wsNormal) {
				TRect boundsRect = BoundsRect;
				boundsRect.left = ini->ReadInteger(
					INI_MAINFORM,
					INI_MAINFORM_LEFT,
					BoundsRect.left);
				boundsRect.top = ini->ReadInteger(
					INI_MAINFORM,
					INI_MAINFORM_TOP,
					BoundsRect.top);
				boundsRect.right = ini->ReadInteger(
					INI_MAINFORM,
					INI_MAINFORM_RIGHT,
					BoundsRect.right);
				boundsRect.bottom = ini->ReadInteger(
					INI_MAINFORM,
					INI_MAINFORM_BOTTOM,
					BoundsRect.bottom);
				BoundsRect = boundsRect;
				try {
					ListView1->Columns->BeginUpdate();
					for(int i = 0; i < ListView1->Columns->Count; i++) {
						ListView1->Columns->Items[i]->Width = ini->ReadInteger(
							INI_MAINFORM,
							INI_MAINFORM_LISTCOLUMNWIDTH + i,
							ListView1->Columns->Items[i]->Width);
					}
				}
				__finally {
					ListView1->Columns->EndUpdate();
				}
			}
		}
		{
			try {
				FileHistory->BeginUpdate();
				FileHistory->Clear();
				AnsiString fileName;
				for(int i = 0; i < FileHistory->Capacity; i++) {
					fileName = ini->ReadString(
						INI_FILEHISTORY,
						INI_FILEHISTORY_ITEM + i,
						"");
					if(fileName.Length() <= 0) {
						break;
					}
					FileHistory->Add(fileName);
				}
			}
			__finally {
				FileHistory->EndUpdate();
			}
			try {
				FileHistory->BeginUpdate();
				FileHistory->Clear();
				AnsiString key;
				for(int i = 0; i < FileHistory->Capacity; i++) {
					key = INI_FILEHISTORY_ITEM + i;
					if(!ini->ValueExists(
						INI_FILEHISTORY,
						key)) {
						break;
					}
					FileHistory->Insert(FileHistory->Count, ini->ReadString(
						INI_FILEHISTORY,
						key,
						""));
				}
			}
			__finally {
				FileHistory->EndUpdate();
			}
		}
		{
			OpenDialogs->InitialDir = ini->ReadString(
				INI_OPENDIALOGS,
				INI_OPENDIALOGS_INITIALDIR,
				"");
			for(int i = 0; i < OpenDialogs->Count; i++) {
				OpenDialogs->Items[i]->FilterIndex = ini->ReadInteger(
					INI_OPENDIALOGS,
					INI_OPENDIALOGS_FILTERINDEX + i,
					1);
			}
		}
		succeeded = true;
	}
	catch(...) {}
	delete ini;
	return succeeded;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SaveSettings()
{
	for(;;) {
		if(TrySaveSettings()) {
			break;
		}
		if(Application->MessageBox(
			"The program cannot save the settings.\r\n\r\nRetry?",
			Application->Title.c_str(),
			MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
			break;
		}
	}
}
//---------------------------------------------------------------------------
bool __fastcall TMainForm::TrySaveSettings()
{
	bool succeeded = false;
	TMemIniFile* ini = NULL;
	try {
		ini = new TMemIniFile(FILE_INI);
		ini->Clear();
		{
			TViewOptions viewOptions = ViewOptions;
			ini->WriteBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_TOOLBAR,
				viewOptions.ToolBar);
			ini->WriteBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_STATUSBAR,
				viewOptions.StatusBar);
			ini->WriteBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_LASTPOSITION,
				viewOptions.LastPosition);
			ini->WriteBool(
				INI_VIEWOPTIONS,
				INI_VIEWOPTIONS_TOPMOST,
				viewOptions.Topmost);
		}
		{
			ini->WriteInteger(
				INI_MAINFORM,
				INI_MAINFORM_WINDOWSTATE,
				WindowState);
			TRect boundsRect = BoundsRect;
			ini->WriteInteger(
				INI_MAINFORM,
				INI_MAINFORM_LEFT,
				boundsRect.left);
			ini->WriteInteger(
				INI_MAINFORM,
				INI_MAINFORM_TOP,
				boundsRect.top);
			ini->WriteInteger(
				INI_MAINFORM,
				INI_MAINFORM_RIGHT,
				boundsRect.right);
			ini->WriteInteger(
				INI_MAINFORM,
				INI_MAINFORM_BOTTOM,
				boundsRect.bottom);
			for(int i = 0; i < ListView1->Columns->Count; i++) {
				ini->WriteInteger(
					INI_MAINFORM,
					INI_MAINFORM_LISTCOLUMNWIDTH + i,
					ListView1->Columns->Items[i]->Width);
			}
		}
		{
			for(int i = 0; i < FileHistory->Count; i++) {
				ini->WriteString(
					INI_FILEHISTORY,
					INI_FILEHISTORY_ITEM + i,
					FileHistory->Items[i]);
			}
		}
		{
			ini->WriteString(
				INI_OPENDIALOGS,
				INI_OPENDIALOGS_INITIALDIR,
				OpenDialogs->InitialDir);
			for(int i = 0; i < OpenDialogs->Count; i++) {
				ini->WriteInteger(
					INI_OPENDIALOGS,
					INI_OPENDIALOGS_FILTERINDEX + i,
					OpenDialogs->Items[i]->FilterIndex);
			}
		}
		ini->UpdateFile();
		succeeded = true;
	}
	catch(...) {}
	delete ini;
	return succeeded;
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
AnsiString __fastcall TMainForm::ResultToStr(int AResult)
{
	switch(AResult) {
		case TCompareItem::RESULT_CREATED: return "Created";
		case TCompareItem::RESULT_MODIFIED: return "Modified";
		case TCompareItem::RESULT_NOT_MODIFIED: return "Not modified";
		case TCompareItem::RESULT_AMBIGUOUS: return "Ambiguous";
		case TCompareItem::RESULT_NONE: return "None";
		default: return "(Unknown)";
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateEnabled(bool AForceAll)
{
	ClearM1->Enabled = HistoryItemM1->Visible;
	int itemCount = ListView1->Items->Count;
	bool hasItems = itemCount > 0;
	if(AForceAll || SaveListAsM1->Enabled != hasItems) {
		SaveListAsM1->Enabled = hasItems;
		PerformToAllM1->Enabled = hasItems;

		PerformToAllT1->Enabled = PerformToAllM1->Enabled;
	}
	int selCount = ListView1->SelCount;
	bool hasSelection = selCount > 0;
	if(AForceAll || SourceM1->Enabled != hasSelection) {
		SourceM1->Enabled = hasSelection;
		DestinationM1->Enabled = hasSelection;
		DeleteM1->Enabled = hasSelection;
		SizeRemaindersM1->Enabled = hasSelection;
		PerformToSelectionM1->Enabled = hasSelection;

		SourceP1->Enabled = SourceM1->Enabled;
		DestinationP1->Enabled = DestinationM1->Enabled;
		DeleteP1->Enabled = DeleteM1->Enabled;
		SizeRemaindersP1->Enabled = SizeRemaindersM1->Enabled;

		CopySourceT1->Enabled = SourceM1->Enabled;
		CopyDestinationT1->Enabled = DestinationM1->Enabled;
		DeleteItemsT1->Enabled = DeleteM1->Enabled;
		PerformToSelectionT1->Enabled = PerformToSelectionM1->Enabled;
	}
	bool hasOneSelection = selCount == 1;
	if(AForceAll || SourceM2->Enabled != hasOneSelection) {
		SourceM2->Enabled = hasOneSelection;
		DestinationM2->Enabled = hasOneSelection &&
			static_cast<TCompareItem*>(ListView1->Selected->Data)->Duplicates;
		BothM1->Enabled = SourceM2->Enabled && DestinationM2->Enabled;
		SourceM3->Enabled = SourceM2->Enabled;
		DestinationM3->Enabled = DestinationM2->Enabled;
		BothM2->Enabled = SourceM3->Enabled && DestinationM3->Enabled;

		SourceP2->Enabled = SourceM2->Enabled;
		DestinationP2->Enabled = DestinationM2->Enabled;
		BothP1->Enabled = BothM1->Enabled;
		SourceP3->Enabled = SourceM3->Enabled;
		DestinationP3->Enabled = DestinationM3->Enabled;
		BothP2->Enabled = BothM2->Enabled;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::UpdateItem(int AIndex)
{
	const TCompareItem& item = *static_cast<TCompareItem*>(
		ListView1->Items->Item[AIndex]->Data);
	{
		ListView1->Items->Item[AIndex]->Caption =
			ExtractFileName(ExtractFileName(item.SrcName));
	}
	{
		TStringList* subItems = new TStringList();
		subItems->Add(ExtractFileDir(item.SrcName));
		subItems->Add(ExtractFileExt(item.SrcName));
		subItems->Add(EnsureSign(FormatInt(item.SizeRem)) + " bytes");
		subItems->Add(item.Duplicates ?
			EnsureSign(FormatInt(item.TimeRemInDays)) + " days" :
			AnsiString("None"));
		subItems->Add(ResultToStr(item.Result));
		ListView1->Items->Item[AIndex]->SubItems->Assign(subItems);
		delete subItems;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ItemCountChange()
{
	ShowStatus(FormatInt(ListView1->Items->Count) + " items are set in total");
	UpdateEnabled(false);
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::SetViewOptions(TViewOptions AViewOptions)
{
	if(AViewOptions != FViewOptions) {
		FViewOptions = AViewOptions;
		ViewOptionsChange();
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ViewOptionsChange()
{
	ToolBar1->Visible = FViewOptions.ToolBar;
	StatusBar1->Visible = FViewOptions.StatusBar;
	SetWindowPos(Handle, FViewOptions.Topmost ? HWND_TOPMOST : HWND_NOTOPMOST,
		0, 0, 0, 0, SWP_NOMOVE | SWP_NOSIZE);

	ToolbarM1->Checked = FViewOptions.ToolBar;
	StatusBarM1->Checked = FViewOptions.StatusBar;
	LastPositionM1->Checked = FViewOptions.LastPosition;
	TopmostM1->Checked = FViewOptions.Topmost;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::FilesDrop(const TStrings* AFiles)
{
	for(;;) {
		TStrings* files = const_cast<TStrings*>(AFiles);
		if(files->Count != 1) {
			Application->MessageBox(
				"Set one item.",
				Application->Title.c_str(),
				MB_OK | MB_ICONERROR | MB_SETFOREGROUND);
			break;
		}
		CompareFiles(files->Strings[0]);
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::WMDropFiles(TWMDropFiles& AMessage)
{
	SetForegroundWindow(Application->Handle);
	TStringList* files = new TStringList();
	{
		HDROP handle = reinterpret_cast<HDROP>(AMessage.Drop);
		int fileCount = DragQueryFile(handle, 0xFFFFFFFF, NULL, 0);
		const int bufferSize = MAX_PATH;
		char buffer[bufferSize];
		for(int i = 0; i < fileCount; i++) {
			if(DragQueryFile(handle, i, buffer, bufferSize) > 0) {
				files->Add(ShortFileNameToLongFileName(buffer));
			}
		}
		DragFinish(handle);
	}
	if(files->Count > 0) {
		FilesDrop(files);
	}
	delete files;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TMainForm::TMainForm(TComponent* AOwner) :
	TForm(AOwner),
	FOpenDialogs(new TOpenDialogList()),
	FFileHistoryMenuItems(new TList()),
	FFileHistory(new TFileHistory),
	FViewOptions()
{
	Application->OnIdle = ApplicationIdle;
	DragAcceptFiles(Handle, true);
	InitOpenDialogs();
	InitFileHistory();
	InitCompareDialog();
	ViewOptionsChange();
	UpdateEnabled(true);
	ResizeListColumnsM1->Click();
	LoadSettings();
	ShowStatus("Ready");
}
//---------------------------------------------------------------------------
__fastcall TMainForm::~TMainForm()
{
	ListView1->Clear();
	UninitCompareDialog();
	UninitFileHistory();
	UninitOpenDialogs();
	delete FFileHistory; FFileHistory = NULL;
	delete FFileHistoryMenuItems; FFileHistoryMenuItems = NULL;
	delete FOpenDialogs; FOpenDialogs = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TMainForm::ShowStatus(AnsiString AStatus)
{
	StatusBar1->SimpleText = AStatus;
}
//---------------------------------------------------------------------------


