//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TMessageDialog.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// class TMessageDialog
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
__fastcall TMessageDialog::TMessageDialog(const TMessageDialog& ASrc) :
	FCaption(),
	FMessage(),
	FButtonCaptions(NULL),
	FDefaultButton(-1),
	FCancelButton(-1),
	FIconType(ICONTYPE_CUSTOM),
	FIcon(NULL),
	FResultButton(-1)
{
	throw 0;
}
//---------------------------------------------------------------------------
TMessageDialog& __fastcall TMessageDialog::operator=(
	const TMessageDialog& ASrc)
{
	throw 0;
}
//---------------------------------------------------------------------------
bool __fastcall TMessageDialog::operator==(const TMessageDialog& ASrc) const
{
	throw 0;
}
//---------------------------------------------------------------------------
bool __fastcall TMessageDialog::operator!=(const TMessageDialog& ASrc) const
{
	throw 0;
}
//---------------------------------------------------------------------------
TForm* __fastcall TMessageDialog::MakeForm() const
{
	TForm* form = CreateMessageDialog(
		Message,
		MakeMsgDlgType(),
		MakeMsgDlgButtons());
	form->Caption = Caption;
	InitButtons(form);
	InitIcon(form);
	form->OnKeyDown = NULL; // forbids Ctrl+C to copy the message
	form->OnCloseQuery = FormCloseQuery;
	return form;
}
//---------------------------------------------------------------------------
TMsgDlgType __fastcall TMessageDialog::MakeMsgDlgType() const
{
	switch(IconType) {
		case ICONTYPE_WARNING: return mtWarning;
		case ICONTYPE_ERROR: return mtError;
		case ICONTYPE_INFORMATION: return mtInformation;
		case ICONTYPE_CONFIRMATION: return mtConfirmation;
		case ICONTYPE_CUSTOM: return mtCustom;
		case ICONTYPE_OTHER: return mtInformation;
		default: return mtCustom;
	}
}
//---------------------------------------------------------------------------
TMsgDlgButtons __fastcall TMessageDialog::MakeMsgDlgButtons() const
{
	const TMsgDlgBtn buttons[] = {
		mbYes,
		mbNo,
		mbOK,
		mbCancel,
		mbAbort,
		mbRetry,
		mbIgnore,
		mbAll,
		mbNoToAll,
		mbYesToAll,
		mbHelp,
	};
	const int buttonCount = sizeof(buttons) / sizeof(TMsgDlgBtn);

	TMsgDlgButtons msgDlgButtons;
	for(int i = 0; i < ButtonCaptions->Count && i < buttonCount; i++) {
		msgDlgButtons << buttons[i];
	}
	return msgDlgButtons;
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::InitButtons(TForm* AForm) const
{
	TList* buttons = new TList();
	EnumControls(buttons, AForm, __classid(TButton));
	buttons->Sort(CompareButton);
	TButton* button;
	for(int i = 0; i < buttons->Count; i++) {
		button = static_cast<TButton*>(buttons->Items[i]);
		button->Caption = ButtonCaptions->Strings[i];
		button->Default = i == DefaultButton;
		if(button->Default) {
			AForm->ActiveControl = button;
		}
		button->Cancel = i == CancelButton;
		button->ModalResult = mrOk;
		button->Tag = i;
		button->OnClick = ButtonClick;
	}
	EnableCloseButton(AForm, CancelButton >= 0);
	delete buttons;
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::EnumControls(TList* AControls, TForm* AForm,
	TClass AClass)
{
	AControls->Clear();
	TControl* control;
	for(int i = 0; i < AForm->ControlCount; i++) {
		control = AForm->Controls[i];
		if(control->InheritsFrom(AClass)) {
			AControls->Add(control);
		}
	}
}
//---------------------------------------------------------------------------
int __fastcall TMessageDialog::CompareButton(void* AItem1, void* AItem2)
{
	return
		static_cast<TButton*>(AItem1)->TabOrder -
		static_cast<TButton*>(AItem2)->TabOrder;
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::EnableCloseButton(TForm* AForm, bool AEnabled)
{
	EnableMenuItem(
		GetSystemMenu(AForm->Handle, false),
		SC_CLOSE,
		MF_BYCOMMAND | (AEnabled ? MF_ENABLED : MF_GRAYED));
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::InitIcon(TForm* AForm) const
{
	if(IconType == ICONTYPE_OTHER) {
		TList* images = new TList();
		EnumControls(images, AForm, __classid(TImage));
		TImage* image;
		for(int i = 0; i < images->Count; i++) {
			image = static_cast<TImage*>(images->Items[i]);
			image->Picture->Icon->Assign(Icon);
		}
		delete images;
	}
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::FormCloseQuery(TObject* AObject,
	bool& ACanClose)
{
	bool canClose = true;
	if(dynamic_cast<TCustomForm*>(AObject)->ModalResult == mrCancel) {
		if(CancelButton >= 0) {
			ResultButton = CancelButton;
		}
		else {
			canClose = false;
		}
	}
	ACanClose = canClose;
}
//---------------------------------------------------------------------------
void __fastcall TMessageDialog::ButtonClick(TObject* AObject)
{
	ResultButton = dynamic_cast<TButton*>(AObject)->Tag;
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TMessageDialog::TMessageDialog() :
	FCaption(Application->Title),
	FMessage(),
	FButtonCaptions(new TStringList()),
	FDefaultButton(-1),
	FCancelButton(-1),
	FIconType(ICONTYPE_CUSTOM),
	FIcon(new TIcon()),
	FResultButton(-1)
{
	// do nothing
}
//---------------------------------------------------------------------------
__fastcall TMessageDialog::~TMessageDialog()
{
	delete FIcon; FIcon = NULL;
	delete FButtonCaptions; FButtonCaptions = NULL;
}
//---------------------------------------------------------------------------
int __fastcall TMessageDialog::Execute()
{
	ResultButton = -1;
	TForm* form = MakeForm(); form->ShowModal(); delete form;
	return ResultButton;
}
//---------------------------------------------------------------------------
