//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <Clipbrd.hpp>
#include "TOperationsDialog.h"
#include "TOperationsThread.h"
//---------------------------------------------------------------------------
#pragma resource "*.dfm"
TOperationsDialog* OperationsDialog = NULL;
//---------------------------------------------------------------------------
// __published:
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::FormShow(TObject* ASender)
{
	ClearLog();
	PercentDone = 0;
	Button1->Enabled = true;
	Button2->Enabled = true;
	Button3->Enabled = true;
	Button4->Enabled = true;
	Button4->Caption = "Cancel";
	ActiveControl = NULL;
	ActiveControl = Button2;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::FormCloseQuery(TObject* ASender,
	bool& ACanClose)
{
	bool canClose = true;
	if(FThread != NULL) {
		FThread->Terminate();
		canClose = false;
	}
	ACanClose = canClose;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::FormClose(TObject* ASender,
	TCloseAction& AAction)
{
	ClearLog();
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::CopySelectionM1Click(TObject* ASender)
{
	TStringList* log = new TStringList();
	for(;;) {
		if(ASender == CopySelectionM1) {
			CopyLogSelection(log);
		}
		else if(ASender == CopyAllM1) {
			CopyLog(log);
		}
		else{
			break;
		}
		Clipboard()->AsText = log->Text;
		break;
	}
	delete log;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::SelectAllM1Click(TObject* ASender)
{
	try {
		ListBox1->Items->BeginUpdate();
		if(ASender == SelectAllM1) {
			ListBox1->SelectAll();
		}
		else if(ASender == SelectNoneM1) {
			ListBox1->ClearSelection();
		}
		else if(ASender == InvertSelectionM1) {
			for(int i = 0; i < ListBox1->Items->Count; i++) {
				ListBox1->Selected[i] = !ListBox1->Selected[i];
			}
		}
	}
	__finally {
		ListBox1->Items->EndUpdate();
	}
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::ButtonClick(TObject* ASender)
{
	for(;;) {
		int operation;
		if(ASender == Button1) {
			operation = TOperationsThread::OPERATION_MOVE;
		}
		else if(ASender == Button2) {
			operation = TOperationsThread::OPERATION_COPY;
		}
		else if(ASender == Button3) {
			if(Application->MessageBox(
				"Deleted files cannot be restored.\r\n\r\nContinue?",
				Application->Title.c_str(),
				MB_OKCANCEL | MB_ICONEXCLAMATION | MB_SETFOREGROUND) != IDOK) {
				break;
			}
			operation = TOperationsThread::OPERATION_DELETE;
		}
		else {
			break;
		}
		ClearLog();
		PercentDone = 0;
		Button1->Enabled = false;
		Button2->Enabled = false;
		Button3->Enabled = false;
		Button4->Enabled = true;
		Button4->Caption = "Cancel";
		Button4->SetFocus();
		TOperationsThread* thread = new TOperationsThread(FFrom, FTo,
			operation);
		thread->OnLog = OperationsThreadLog;
		thread->OnProgress = OperationsThreadProgress;
		thread->OnTerminate = OperationsThreadTerminate;
		FThread = thread;
		FThread->Resume();
		break;
	}
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::OperationsThreadLog(TObject* ASender,
	AnsiString ALog)
{
	AddLog(ALog);
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::OperationsThreadProgress(TObject* ASender,
	int AProcessed, int AAll)
{
	ShowProgress(AProcessed, AAll);
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::OperationsThreadTerminate(TObject* ASender)
{
	PercentDone = 0;
	Button1->Enabled = false;
	Button2->Enabled = false;
	Button3->Enabled = false;
	Button4->Enabled = true;
	Button4->Caption = "OK";
	Button4->SetFocus();
	FThread = NULL;
}
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
int __fastcall TOperationsDialog::GetPercentDone() const
{
	return ProgressBar1->Position;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::SetPercentDone(int APercentDone)
{
	ProgressBar1->Position = APercentDone;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TOperationsDialog::TOperationsDialog(TComponent* AOwner) :
	TForm(AOwner),
	FFrom(NULL),
	FTo(NULL),
	FThread(NULL)
{
	// do nothing
}
//---------------------------------------------------------------------------
__fastcall TOperationsDialog::~TOperationsDialog()
{
	// do nothing
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::Execute(const TStrings* AFrom,
	const TStrings* ATo)
{
	FFrom = AFrom;
	FTo = ATo;
 	ShowModal();
 	FFrom = FTo = NULL;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::AddLog(AnsiString ALog)
{
	ListBox1->Items->Add(ALog);
	int width = ListBox1->Canvas->TextWidth(ALog);
	if(width > ListBox1->ScrollWidth) {
		ListBox1->ScrollWidth = width;
	}
	if(ListBox1->Count > 0) {
		ListBox1->ItemIndex = ListBox1->Items->Count - 1;
	}
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::ClearLog()
{
	ListBox1->Items->Clear();
	ListBox1->ScrollWidth = 0;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::CopyLog(TStrings* ALog)
{
	ALog->Assign(ListBox1->Items);
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::CopyLogSelection(TStrings* ALog)
{
	ALog->Clear();
	for(int i = 0; i < ListBox1->Items->Count; i++) {
		if(ListBox1->Selected[i]) {
			ALog->Add(ListBox1->Items->Strings[i]);
		}
	}
}
//---------------------------------------------------------------------------
void __fastcall TOperationsDialog::ShowProgress(int AProcessed, int AAll)
{
	int percentDone;
	if(AProcessed <= 0 || AAll <= 0) {
		percentDone = 0;
	}
	else if(AProcessed >= AAll) {
		percentDone = 100;
	}
	else {
		percentDone = static_cast<__int64>(AProcessed) * 100 / AAll;
	}
	PercentDone = percentDone;
}
//---------------------------------------------------------------------------

