//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "TOperationsThread.h"
#include "MySysUtils.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
// class TOperationsThread : public TReportThread
//---------------------------------------------------------------------------
// private:
//---------------------------------------------------------------------------
AnsiString __fastcall TOperationsThread::DoubleQuoteStr(AnsiString AStr)
{
	return "\"" + AStr + "\"";
}
//---------------------------------------------------------------------------
AnsiString __fastcall TOperationsThread::ResultToStr(bool ASucceeded)
{
	return ASucceeded ? "Succeeded" : "Not succeeded";
}
//---------------------------------------------------------------------------
AnsiString __fastcall TOperationsThread::DoneToStr(int AProcessed, int ACount)
{
	return AProcessed >= ACount ?
		"Done" :
		(AProcessed > 0 ? "Stopped" : "Canceled");
}
//---------------------------------------------------------------------------
bool __fastcall TOperationsThread::PerformFileOperation(AnsiString AFrom,
	AnsiString ATo, bool ACopy, bool ADelete)
{
	bool succeeded = false;
	for(;;) {
		if(ACopy) {
			if(!ForceDirectories(ExtractFileDir(ATo))) {
				break;
			}
			if(!::CopyFile(AFrom.c_str(), ATo.c_str(), false)) {
				break;
			}
		}
		if(ADelete) {
			if(!::DeleteFile(AFrom.c_str())) {
				break;
			}
		}
		succeeded = true;
		break;
	}
	return succeeded;
}
//---------------------------------------------------------------------------
void __fastcall TOperationsThread::PerformOperations()
{
	int processed = 0;
	int succeeded = 0;
	int count = FFrom->Count;
	for(int i = 0; i < count; i++) {
		if(Terminated) {
			break;
		}
		switch(FOperation) {
			case OPERATION_MOVE:
				if(PerformMove(FFrom->Strings[i], FTo->Strings[i])) {
					succeeded++;
				}
				break;
			case OPERATION_COPY:
				if(PerformCopy(FFrom->Strings[i], FTo->Strings[i])) {
					succeeded++;
				}
				break;
			case OPERATION_DELETE:
				if(PerformDelete(FFrom->Strings[i])) {
					succeeded++;
				}
				break;
		}
		DoProgress(++processed, count);
	}
	DoLog(DoneToStr(processed, count));
	DoLog("Succeeded: " + FormatInt(succeeded) + " items");
	DoLog("Not succeeded: " + FormatInt(processed - succeeded) + " items");
}
//---------------------------------------------------------------------------
bool __fastcall TOperationsThread::PerformMove(AnsiString AFrom,
	AnsiString ATo)
{
	DoLog("Moving " + DoubleQuoteStr(ExtractFileName(AFrom)) + " from " +
		DoubleQuoteStr(ExtractFileDir(AFrom)) + " to " +
		DoubleQuoteStr(ExtractFileDir(ATo)));
	bool succeeded = PerformFileOperation(AFrom, ATo, true, true);
	DoLog(ResultToStr(succeeded));
	return succeeded;
}
//---------------------------------------------------------------------------
bool __fastcall TOperationsThread::PerformCopy(AnsiString AFrom,
	AnsiString ATo)
{
	DoLog("Copying " + DoubleQuoteStr(ExtractFileName(AFrom)) + " from " +
		DoubleQuoteStr(ExtractFileDir(AFrom)) + " to " +
		DoubleQuoteStr(ExtractFileDir(ATo)));
	bool succeeded = PerformFileOperation(AFrom, ATo, true, false);
	DoLog(ResultToStr(succeeded));
	return succeeded;
}
//---------------------------------------------------------------------------
bool __fastcall TOperationsThread::PerformDelete(AnsiString ATarget)
{
	DoLog("Deleting " + DoubleQuoteStr(ExtractFileName(ATarget)) + " in " +
		DoubleQuoteStr(ExtractFileDir(ATarget)));
	bool succeeded = PerformFileOperation(ATarget, "", false, true);
	DoLog(ResultToStr(succeeded));
	return succeeded;
}
//---------------------------------------------------------------------------
// protected:
//---------------------------------------------------------------------------
void __fastcall TOperationsThread::Execute()
{
	TReportThread::Execute();
	if(Error) {
		DoLog("Error");
	}
}
//---------------------------------------------------------------------------
void __fastcall TOperationsThread::Prepare()
{
	// do nothing
}
//---------------------------------------------------------------------------
void __fastcall TOperationsThread::Process()
{
	PerformOperations();
}
//---------------------------------------------------------------------------
void __fastcall TOperationsThread::Unprepare()
{
	delete FTo; FTo = NULL;
	delete FFrom; FFrom = NULL;
}
//---------------------------------------------------------------------------
// public:
//---------------------------------------------------------------------------
__fastcall TOperationsThread::TOperationsThread(const TStrings* AFrom,
	const TStrings* ATo, int AOperation) :
	TReportThread(true),
	FFrom(new TStringList()),
	FTo(new TStringList()),
	FOperation(AOperation)
{
	FreeOnTerminate = true;
	FFrom->Assign(const_cast<TStrings*>(AFrom));
	FTo->Assign(const_cast<TStrings*>(ATo));
}
//---------------------------------------------------------------------------

